@@ -95,16 +95,17 @@ When properly installed, the on-disk layout of an app managed by esky looks
 like this:
 
     prog.exe                     - esky bootstrapping executable
-    appname-X.Y.platform/        - specific version of the application
-      prog.exe                   - executable(s) as produced by freezer module
-      library.zip                - pure-python frozen modules
-      pythonXY.dll               - python DLL
-      esky-files/                - esky control files
-        bootstrap/               - files not yet moved into bootstrapping env
-        bootstrap-manifest.txt   - list of files expected in bootstrap env
-        lockfile.txt             - lockfile to block removal of in-use versions
-      ...other deps...
-    updates/                     - work area for fetching/unpacking updates
+    appdata/                     - container for all the esky magic
+      appname-X.Y.platform/      - specific version of the application
+        prog.exe                 - executable(s) as produced by freezer module
+        library.zip              - pure-python frozen modules
+        pythonXY.dll             - python DLL
+        esky-files/              - esky control files
+          bootstrap/             - files not yet moved into bootstrapping env
+          bootstrap-manifest.txt - list of files expected in bootstrap env
+          lockfile.txt           - lockfile to block removal of in-use versions
+        ...other deps...
+      updates/                   - work area for fetching/unpacking updates
 
 This is also the layout of the zipfiles produced by bdist_esky.  The 
 "appname-X.Y" directory is simply a frozen app directory with some extra
@@ -136,8 +137,8 @@ from __future__ import with_statement
 from __future__ import absolute_import
 
 __ver_major__ = 0
-__ver_minor__ = 8
-__ver_patch__ = 5
+__ver_minor__ = 9
+__ver_patch__ = 0
 __ver_sub__ = ""
 __ver_tuple__ = (__ver_major__,__ver_minor__,__ver_patch__,__ver_sub__)
 __version__ = "%d.%d.%d%s" % __ver_tuple__
@@ -154,7 +155,9 @@ from esky.util import split_app_version, join_app_version,\
                       is_version_dir, is_uninstalled_version_dir,\
                       parse_version, get_best_version, appdir_from_executable,\
                       copy_ownership_info, lock_version_dir, ESKY_CONTROL_DIR,\
-                      files_differ, lazy_import
+                      files_differ, lazy_import, ESKY_APPDATA_DIR, \
+                      get_all_versions, is_locked_version_dir, \
+                      is_installed_version_dir, really_rmtree, really_rename
 
 #  Since all frozen apps are required to import this module and call the
 #  run_startup_hooks() function, we use a simple lazy import mechanism to 
@@ -165,11 +168,6 @@ def os():
     import os
     return os
 
-@lazy_import
-def shutil():
-    import shutil
-    return shutil
-
 @lazy_import
 def socket():
     import socket
@@ -259,12 +257,14 @@ class Esky(object):
         """
         if os.path.isfile(appdir_or_exe):
             self.appdir = appdir_from_executable(appdir_or_exe)
-            vdir = appdir_or_exe[len(self.appdir):].split(os.sep)[1]
+            vsdir = self._get_versions_dir()
+            vdir = appdir_or_exe[len(vsdir):].split(os.sep)[1]
             details = split_app_version(vdir)
             self.name,self.active_version,self.platform = details
         else:
             self.active_version = None
             self.appdir = appdir_or_exe
+        self.appdir = os.path.abspath(self.appdir)
 
     def _get_version_finder(self):
         return self.__version_finder
@@ -278,7 +278,21 @@ class Esky(object):
 
     def _get_update_dir(self):
         """Get the directory path in which self.version_finder can work."""
-        return os.path.join(self.appdir,"updates")
+        return os.path.join(self._get_versions_dir(),"updates")
+
+    def _get_versions_dir(self):
+        """Get the directory path containing individual version dirs."""
+        if not ESKY_APPDATA_DIR:
+            return self.appdir
+        # TODO: remove compatability hooks for ESKY_APPDATA_DIR=""
+        try:
+            for nm in os.listdir(os.path.join(self.appdir,ESKY_APPDATA_DIR)):
+                fullnm = os.path.join(self.appdir,ESKY_APPDATA_DIR,nm)
+                if is_version_dir(fullnm) and is_installed_version_dir(fullnm):
+                    return os.path.join(self.appdir,ESKY_APPDATA_DIR)
+        except EnvironmentError:
+            pass
+        return self.appdir
 
     def get_abspath(self,relpath):
         """Get the absolute path of a file within the current version."""
@@ -286,7 +300,10 @@ class Esky(object):
             v = join_app_version(self.name,self.active_version,self.platform)
         else:
             v = join_app_version(self.name,self.version,self.platform)
-        return os.path.abspath(os.path.join(self.appdir,v,relpath))
+        # TODO: remove compatability hooks for ESKY_APPDATA_DIR=""
+        if os.path.exists(os.path.join(self._get_versions_dir(),v)):
+            return os.path.join(self._get_versions_dir(),v,relpath)
+        return os.path.join(self.appdir,v,relpath)
 
     def reinitialize(self):
         """Reinitialize internal state by poking around in the app directory.
@@ -295,7 +312,7 @@ class Esky(object):
         EskyBrokenError will be raised.  This should never happen unless
         another process has been messing with the files.
         """
-        best_version = get_best_version(self.appdir)
+        best_version = get_best_version(self._get_versions_dir())
         if best_version is None:
             raise EskyBrokenError("no frozen versions found")
         details = split_app_version(best_version)
@@ -350,7 +367,7 @@ class Esky(object):
                     if mtime > newest_mtime:
                         newest_mtime = mtime
                 if newest_mtime + self.lock_timeout < time.time():
-                    shutil.rmtree(lockdir)
+                    really_rmtree(lockdir)
                     return self.lock(num_retries+1)
                 else:
                     raise EskyLockedError
@@ -479,8 +496,9 @@ class Esky(object):
         is required without duplicating the logic.
         """
         appdir = self.appdir
-        best_version = get_best_version(appdir)
-        new_version = get_best_version(appdir,include_partial_installs=True)
+        vsdir = self._get_versions_dir()
+        best_version = get_best_version(vsdir)
+        new_version = get_best_version(vsdir,include_partial_installs=True)
         #  If there's a partial install we must complete it, since it
         #  could have left exes in the bootstrap env and we don't want
         #  to accidentally delete their dependencies.
@@ -488,6 +506,15 @@ class Esky(object):
             (_,v,_) = split_app_version(new_version)
             yield (self.install_version,(v,))
             best_version = new_version
+        #  TODO: remove compatability hooks for ESKY_APPDATA_DIR=""
+        if vsdir == appdir and ESKY_APPDATA_DIR:
+            appdatadir = os.path.join(appdir,ESKY_APPDATA_DIR)
+            if os.path.isdir(appdatadir) and os.listdir(appdatadir):
+                new_version = get_best_version(appdatadir,include_partial_installs=True)
+                if best_version != new_version:
+                    (_,v,_) = split_app_version(new_version)
+                    yield (self.install_version,(v,))
+                    best_version = new_version
         #  Now we can safely remove all the old versions.
         #  We except the currently-executing version, and silently
         #  ignore any locked versions.
@@ -495,34 +522,47 @@ class Esky(object):
         manifest.add("updates")
         manifest.add("locked")
         manifest.add(best_version)
-        if self.active_version and self.active_version != best_version:
-            yield lambda: False
+        if self.active_version:
+            if self.active_version != split_app_version(best_version)[1]:
+                yield lambda: False
             manifest.add(self.active_version)
-        for nm in os.listdir(appdir):
-            if nm not in manifest:
-                fullnm = os.path.join(appdir,nm)
-                if is_version_dir(fullnm):
-                    #  It's an installed-but-obsolete version.  Properly
-                    #  uninstall it so it will clean up the bootstrap env.
-                    (_,v,_) = split_app_version(nm)
-                    try:
-                        yield (self.uninstall_version, (v,))
-                    except VersionLockedError:
-                        yield lambda: False
+        # TODO: remove compatability hooks for ESKY_APPDATA_DIR=""
+        for tdir in (appdir,vsdir):
+            for nm in os.listdir(tdir):
+                if nm not in manifest:
+                    fullnm = os.path.join(tdir,nm)
+                    if not os.path.isdir(fullnm):
+                        #  It's an unaccounted-for file in the bootstrap env.
+                        #  Leave it alone.
+                        pass
+                    elif is_version_dir(fullnm):
+                        #  It's an installed-but-obsolete version.  Properly
+                        #  uninstall it so it will clean up the bootstrap env.
+                        (_,v,_) = split_app_version(nm)
+                        try:
+                            yield (self.uninstall_version, (v,))
+                        except VersionLockedError:
+                            yield lambda: False
+                        else:
+                            yield (self._try_remove, (tdir,nm,manifest,))
+                    elif is_uninstalled_version_dir(fullnm):
+                        #  It's a partially-removed version; finish removing it.
+                        yield (self._try_remove, (tdir,nm,manifest,))
+                    elif ".old." in nm or nm.endswith(".old"):
+                        #  It's a temporary backup file; remove it.
+                        yield (self._try_remove, (tdir,nm,manifest,))
                     else:
-                        yield (self._try_remove, (appdir,nm,manifest,))
-                elif is_uninstalled_version_dir(fullnm):
-                    #  It's a partially-removed version; finish removing it.
-                    yield (self._try_remove, (appdir,nm,manifest,))
-                elif ".old." in nm or nm.endswith(".old"):
-                    #  It's a temporary backup file; remove it.
-                    yield (self._try_remove, (appdir,nm,manifest,))
-                else:
-                    #  It's an unaccounted-for entry in the bootstrap env.
-                    #  Can't prove it's safe to remove, so leave it.
-                    pass
+                        for (_,_,filenms) in os.walk(fullnm):
+                            if filenms:
+                                #  It contains unaccounted-for files in the
+                                #  bootstrap env. Can't prove it's safe to
+                                #  remove, so leave it alone.
+                                break
+                        else:
+                            #  It's an empty directory structure, remove it.
+                            yield (self._try_remove, (tdir,nm,manifest,))
         #  If there are pending overwrites, try to do them.
-        ovrdir = os.path.join(appdir,best_version,ESKY_CONTROL_DIR,"overwrite")
+        ovrdir = os.path.join(vsdir,best_version,ESKY_CONTROL_DIR,"overwrite")
         if os.path.exists(ovrdir):
             try:
                 for (dirnm,_,filenms) in os.walk(ovrdir,topdown=False):
@@ -567,11 +607,11 @@ class Esky(object):
             exe = [sys.executable,"-c","import esky; esky.run_startup_hooks()","--esky-spawn-cleanup"]
         else:
             exe = sys.executable
-            #  Try to re-launch the best available version, so that the
-            #  currently in-use version can be cleaned up.
+            #  Try to re-launch the best available version, so that
+            #  the currently in-use version can be cleaned up.
             if self.active_version is not None:
-               appdir = self.appdir
-               bestver = get_best_version(appdir,include_partial_installs=True)
+               vsdir = self._get_versions_dir()
+               bestver = get_best_version(vsdir,include_partial_installs=True)
                if bestver is not None:
                    (_,version,_) = split_app_version(bestver)
                    if self.active_version != version:
@@ -601,8 +641,8 @@ class Esky(object):
             subprocess.Popen(exe,**kwds)
         
 
-    def _try_remove(self,appdir,path,manifest=[]):
-        """Try to remove the file/directory at the given path in the appdir.
+    def _try_remove(self,tdir,path,manifest=[]):
+        """Try to remove the file/directory at given path in the target dir.
 
         This method attempts to remove the file or directory at the given path,
         but will fail silently under a number of conditions:
@@ -613,7 +653,7 @@ class Esky(object):
             * if the path appears in the given manifest
 
         """
-        fullpath = os.path.join(appdir,path)
+        fullpath = os.path.join(tdir,path)
         if fullpath in sys.path:
             return False
         if path in manifest:
@@ -629,11 +669,11 @@ class Esky(object):
                         esky_paths.append(nm)
                     else:
                         subdir = os.path.join(path,nm)
-                        success &= self._try_remove(appdir,subdir,manifest)
+                        success &= self._try_remove(tdir,subdir,manifest)
                 if not success:
                     return False
                 for nm in sorted(esky_paths):
-                    self._try_remove(appdir,os.path.join(path,nm),manifest)
+                    self._try_remove(tdir,os.path.join(path,nm),manifest)
                 os.rmdir(fullpath)
             else:
                 os.unlink(fullpath)
@@ -704,7 +744,7 @@ class Esky(object):
                     cleaned = self.cleanup()
         finally:
             #  Drop root privileges as soon as possible.
-            if not cleaned:
+            if not cleaned and self.needs_cleanup():
                 self.cleanup_at_exit()
             if got_root:
                 self.drop_root()
@@ -749,16 +789,17 @@ class Esky(object):
         if self.version_finder is None:
             raise NoVersionFinderError
         #  Guard against malicious input (might be called with root privs)
+        vsdir = self._get_versions_dir()
         target = join_app_version(self.name,version,self.platform)
-        target = os.path.join(self.appdir,target)
-        assert os.path.dirname(target) == self.appdir
+        target = os.path.join(vsdir,target)
+        assert os.path.dirname(target) == vsdir
         #  Get the new version using the VersionFinder
         loc = self.version_finder.has_version(self,version)
         if not loc:
             loc = self.version_finder.fetch_version(self,version,callback)
         #  Adjust permissions to match the current version
         vdir = join_app_version(self.name,self.version,self.platform)
-        copy_ownership_info(os.path.join(self.appdir,vdir),loc)
+        copy_ownership_info(os.path.join(vsdir,vdir),loc)
         return loc
 
     @allow_from_sudo(str,iterator=True)
@@ -771,9 +812,10 @@ class Esky(object):
         if self.version_finder is None:
             raise NoVersionFinderError
         #  Guard against malicious input (might be called with root privs)
+        vsdir = self._get_versions_dir()
         target = join_app_version(self.name,version,self.platform)
-        target = os.path.join(self.appdir,target)
-        assert os.path.dirname(target) == self.appdir
+        target = os.path.join(vsdir,target)
+        assert os.path.dirname(target) == vsdir
         #  Get the new version using the VersionFinder
         loc = self.version_finder.has_version(self,version)
         if not loc:
@@ -784,7 +826,7 @@ class Esky(object):
                     loc = status["path"]
         #  Adjust permissions to match the current version
         vdir = join_app_version(self.name,self.version,self.platform)
-        copy_ownership_info(os.path.join(self.appdir,vdir),loc)
+        copy_ownership_info(os.path.join(vsdir,vdir),loc)
         yield {"status":"ready","path":loc}
 
     @allow_from_sudo(str)
@@ -798,20 +840,34 @@ class Esky(object):
         if self.sudo_proxy is not None:
             return self.sudo_proxy.install_version(version)
         #  Extract update then rename into position in main app directory
+        vsdir = self._get_versions_dir()
         target = join_app_version(self.name,version,self.platform)
-        target = os.path.join(self.appdir,target)
+        target = os.path.join(vsdir,target)
         #  Guard against malicious input (might be called with root privs)
-        assert os.path.dirname(target) == self.appdir
+        assert os.path.dirname(target) == vsdir
         if not os.path.exists(target):
             self.fetch_version(version)
             source = self.version_finder.has_version(self,version)
+        #  TODO: remove compatability hooks for ESKY_APPDATA_DIR="".
+        #  This is our chance to migrate to the new appdata dir layout,
+        #  by installing into it.
+        if vsdir == self.appdir and ESKY_APPDATA_DIR:
+            vsdir = os.path.join(self.appdir,ESKY_APPDATA_DIR)
+            try:
+                os.mkdir(vsdir)
+            except EnvironmentError, e:
+                if e.errno not in (errno.EEXIST,):
+                    raise
+            else:
+                copy_ownership_info(self.appdir,vsdir)
+            target = os.path.join(vsdir,os.path.basename(target))
         self.lock()
         try:
             if not os.path.exists(target):
-                os.rename(source,target)
+                really_rename(source,target)
             trn = esky.fstransact.FSTransaction(self.appdir)
             try:
-                self._unpack_bootstrap_env(version,trn)
+                self._unpack_bootstrap_env(target,trn)
             except Exception:
                 trn.abort()
                 raise
@@ -820,11 +876,9 @@ class Esky(object):
         finally:
             self.unlock()
 
-    def _unpack_bootstrap_env(self,version,trn):
+    def _unpack_bootstrap_env(self,target,trn):
         """Unpack the bootstrap env from the given target directory."""
-        vdir = join_app_version(self.name,version,self.platform)
-        target = os.path.join(self.appdir,vdir)
-        assert os.path.dirname(target) == self.appdir
+        vdir = os.path.basename(target)
         #  Move new bootrapping environment into main app dir.
         #  Be sure to move dependencies before executables.
         bootstrap = os.path.join(target,ESKY_CONTROL_DIR,"bootstrap")
@@ -839,7 +893,8 @@ class Esky(object):
                     if not files_differ(bssrc,bsdst):
                         trn.remove(bssrc)
                     elif esky.winres.is_safe_to_overwrite(bssrc,bsdst):
-                        ovrdir = os.path.join(target,ESKY_CONTROL_DIR,"overwrite")
+                        ovrdir = os.path.join(target,ESKY_CONTROL_DIR)
+                        ovrdir = os.path.join(ovrdir,"overwrite")
                         if not os.path.exists(ovrdir):
                             os.mkdir(ovrdir)
                         trn.move(bssrc,os.path.join(ovrdir,nm))
@@ -858,19 +913,20 @@ class Esky(object):
         """Uninstall the specified version of the app."""
         if self.sudo_proxy is not None:
             return self.sudo_proxy.uninstall_version(version)
+        vsdir = self._get_versions_dir()
         target_name = join_app_version(self.name,version,self.platform)
-        target = os.path.join(self.appdir,target_name)
+        target = os.path.join(vsdir,target_name)
         #  Guard against malicious input (might be called with root privs)
-        assert os.path.dirname(target) == self.appdir
-        # TODO: remove compatability hooks
+        assert os.path.dirname(target) == vsdir
+        #  TODO: remove compatability hooks for ESKY_APPDATA_DIR="".
+        if ESKY_APPDATA_DIR and not os.path.exists(target):
+            if vsdir == self.appdir:
+                target = os.path.join(self.appdir,ESKY_APPDATA_DIR,target_name)
+            else:
+                target = os.path.join(self.appdir,target_name)
         lockfile = os.path.join(target,ESKY_CONTROL_DIR,"lockfile.txt")
-        if not os.path.exists(lockfile):
-            lockfile = os.path.join(target,"esky-lockfile.txt")
         bsfile = os.path.join(target,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")
         bsfile_old = os.path.join(target,ESKY_CONTROL_DIR,"bootstrap-manifest-old.txt")
-        if not os.path.exists(bsfile):
-            bsfile = os.path.join(target,"esky-bootstrap.txt")
-            bsfile_old = os.path.join(target,"esky-bootstrap-old.txt")
         self.lock()
         try:
             if not os.path.exists(target):
@@ -894,7 +950,7 @@ class Esky(object):
             #  To avoid clobbering in-use version, respect locks on this file.
             if sys.platform == "win32":
                 try:
-                    os.rename(bsfile,bsfile_old)
+                    really_rename(bsfile,bsfile_old)
                 except EnvironmentError:
                     raise VersionLockedError("version in use: %s" % (version,))
             else:
@@ -912,7 +968,7 @@ class Esky(object):
                         msg = "version in use: %s" % (version,)
                         raise VersionLockedError(msg)
                     else:
-                        os.rename(bsfile,bsfile_old)
+                        really_rename(bsfile,bsfile_old)
                     finally:
                         f.close()
         finally:
@@ -923,7 +979,7 @@ class Esky(object):
         target_name = join_app_version(self.name,version,self.platform)
         #  Get set of all files that must stay in the main appdir
         to_keep = set()
-        for vname in os.listdir(self.appdir):
+        for vname in os.listdir(self._get_versions_dir()):
             if vname == target_name:
                 continue
             details = split_app_version(vname)
@@ -948,9 +1004,16 @@ class Esky(object):
         This is the set of files/directories that the given version expects
         to be in the main app directory.
         """
-        mpath = os.path.join(self.appdir,vdir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")
+        vsdir = self._get_versions_dir()
+        mpath = os.path.join(vsdir,vdir,ESKY_CONTROL_DIR)
+        mpath = os.path.join(mpath,"bootstrap-manifest.txt")
+        #  TODO: remove compatability hooks for ESKY_APPDATA_DIR="".
         if not os.path.exists(mpath):
-            mpath = os.path.join(self.appdir,vdir,"esky-bootstrap.txt")
+            if vsdir == self.appdir:
+                mpath = os.path.join(self.appdir,ESKY_APPDATA_DIR,vdir,ESKY_CONTROL_DIR)
+            else:
+                mpath = os.path.join(self.appdir,vdir,ESKY_CONTROL_DIR)
+            mpath = os.path.join(mpath,"bootstrap-manifest.txt")
         manifest = set()
         try:
             with open(mpath,"rt") as mf:
@@ -976,14 +1039,24 @@ def run_startup_hooks():
     # delete files out from under us.
     if getattr(sys,"frozen",False):
         appdir = appdir_from_executable(sys.executable)
-        vdir = sys.executable[len(appdir):].split(os.sep)[1]
-        vdir = os.path.join(appdir,vdir)
-        if is_version_dir(vdir):
-            lock_version_dir(vdir)
+        # TODO: remove ESKY_APPDATA_DIR="" compatability hooks
+        if ESKY_APPDATA_DIR:
+            vdir = os.sep.join(sys.executable[len(appdir):].split(os.sep)[1:3])
+            vdir = os.path.join(appdir,vdir)
+            if not is_version_dir(vdir):
+                vdir = sys.executable[len(appdir):].split(os.sep)[1]
+                vdir = os.path.join(appdir,vdir)
+        else:
+            vdir = sys.executable[len(appdir):].split(os.sep)[1]
+            vdir = os.path.join(appdir,vdir)
+            if not is_version_dir(vdir):
+                vdir = os.sep.join(sys.executable[len(appdir):].split(os.sep)[1:3])
+                vdir = os.path.join(appdir,vdir)
+        lock_version_dir(vdir)
     # Run the "spawn-cleanup" hook if given.
     if len(sys.argv) > 1 and sys.argv[1] == "--esky-spawn-cleanup":
         app = pickle.loads(base64.b64decode(sys.argv[2]))
-        time.sleep(10)        
+        time.sleep(1)        
         app.cleanup()
         sys.exit(0)
     # Let esky.slaveproc run its hooks.
