@@ -22,6 +22,7 @@ import sys
 import shutil
 import zipfile
 import tempfile
+import hashlib
 import inspect
 from glob import glob
 
@@ -31,11 +32,24 @@ from distutils.util import convert_path
 
 import esky.patch
 from esky.util import get_platform, is_core_dependency, create_zipfile, \
-                      split_app_version, join_app_version, ESKY_CONTROL_DIR
+                      split_app_version, join_app_version, ESKY_CONTROL_DIR, \
+                      ESKY_APPDATA_DIR, really_rmtree
+
 if sys.platform == "win32":
     from esky import winres
     from xml.dom import minidom
 
+try:
+    from esky.bdist_esky import pypyc
+except ImportError:
+    pypyc = None
+    COMPILED_BOOTSTRAP_CACHE = None
+else:
+    COMPILED_BOOTSTRAP_CACHE = os.path.dirname(__file__)
+    if not os.path.isdir(COMPILED_BOOTSTRAP_CACHE):
+        COMPILED_BOOTSTRAP_CACHE = None
+
+
 
 #  setuptools likes to be imported before anything else that
 #  might monkey-patch distutils.  We don't actually use it,
@@ -116,6 +130,8 @@ class Executable(unicode):
         if self._name is not None:
             nm = self._name
         else:
+            if not isinstance(self.script,basestring):
+                raise TypeError("Must specify name if script is not a file")
             nm = os.path.basename(self.script)
             if nm.endswith(".py"):
                 nm = nm[:-3]
@@ -128,7 +144,10 @@ class Executable(unicode):
     @property
     def gui_only(self):
         if self._gui_only is None:
-            return self.script.endswith(".pyw")
+            if not isinstance(self.script,basestring):
+                return False
+            else:
+                return self.script.endswith(".pyw")
         else:
             return self._gui_only
 
@@ -162,13 +181,19 @@ class bdist_esky(Command):
                          freezer module.
 
         bootstrap_module:  a custom module to use for esky bootstrapping;
-                           the default calls esky.bootstrap.common.bootstrap().
+                           the default calls esky.bootstrap.bootstrap()
 
         bootstrap_code:  a custom code string to use for esky bootstrapping;
                          this precludes the use of the bootstrap_module option.
                          If a non-string object is given, its source is taken
                          using inspect.getsource().
 
+        compile_bootstrap_exes:  whether to compile the bootstrapping code to a
+                                 stand-alone exe; this requires PyPy installed
+                                 and the bootstrap code to be valid RPython.
+                                 When false, the bootstrap env will use a
+                                 trimmed-down copy of the freezer module exe.
+
         dont_run_startup_hooks:  don't force all executables to call
                                  esky.run_startup_hooks() on startup.
 
@@ -202,6 +227,8 @@ class bdist_esky(Command):
                      "module to use for bootstrapping the application"),
                     ('bootstrap-code=', None,
                      "code to use for bootstrapping the application"),
+                    ('compile-bootstrap-exes=', None,
+                     "whether to compile the bootstrapping exes with pypy"),
                     ('bundle-msvcrt=', None,
                      "whether to bundle MSVCRT as private assembly"),
                     ('includes=', None,
@@ -214,9 +241,11 @@ class bdist_esky(Command):
                      "function to call just before starting to freeze the app"),
                     ('pre-zip-callback=', None,
                      "function to call just before starting to zip up the app"),
+                    ('enable-appdata-dir=', None,
+                     "enable new 'appdata' directory layout (will go away after the 0.9.X series)"),
                    ]
 
-    boolean_options = ["bundle-msvcrt","dont-run-startup-hooks"]
+    boolean_options = ["bundle-msvcrt","dont-run-startup-hooks","compile-bootstrap-exes","enable-appdata-dir"]
 
     def initialize_options(self):
         self.dist_dir = None
@@ -228,11 +257,17 @@ class bdist_esky(Command):
         self.dont_run_startup_hooks = False
         self.bootstrap_module = None
         self.bootstrap_code = None
+        self.compile_bootstrap_exes = False
+        self._compiled_exes = {}
         self.pre_freeze_callback = None
         self.pre_zip_callback = None
+        self.enable_appdata_dir = False
 
     def finalize_options(self):
         self.set_undefined_options('bdist',('dist_dir', 'dist_dir'))
+        if self.compile_bootstrap_exes and pypyc is None:
+            err = "compiling bootstrap exes requires pypy installed."
+            raise RuntimeError(err)
         if self.freezer_module is None:
             for freezer_module in ("py2exe","py2app","bbfreeze","cxfreeze"):
                 self.freezer_module = _FREEZERS[freezer_module]
@@ -273,7 +308,7 @@ class bdist_esky(Command):
         try:
             self._run()
         finally:
-            shutil.rmtree(self.tempdir)
+            really_rmtree(self.tempdir)
 
     def _run(self):
         self._run_initialise_dirs()
@@ -290,10 +325,14 @@ class bdist_esky(Command):
         platform = get_platform()
         self.bootstrap_dir = os.path.join(self.dist_dir,
                                           "%s.%s"%(fullname,platform,))
-        self.freeze_dir = os.path.join(self.bootstrap_dir,
-                                       "%s.%s"%(fullname,platform,))
+        if self.enable_appdata_dir:
+            self.freeze_dir = os.path.join(self.bootstrap_dir,ESKY_APPDATA_DIR,
+                                           "%s.%s"%(fullname,platform,))
+        else:
+            self.freeze_dir = os.path.join(self.bootstrap_dir,
+                                           "%s.%s"%(fullname,platform,))
         if os.path.exists(self.bootstrap_dir):
-            shutil.rmtree(self.bootstrap_dir)
+            really_rmtree(self.bootstrap_dir)
         os.makedirs(self.freeze_dir)
 
     def _run_freeze_scripts(self):
@@ -305,10 +344,6 @@ class bdist_esky(Command):
             lockfile = os.path.join(self.freeze_dir,ESKY_CONTROL_DIR,"lockfile.txt")
             with open(lockfile,"w") as lf:
                 lf.write("this file is used by esky to lock the version dir\n")
-            # TODO: remove compatability hook
-            shutil.copyfile(os.path.join(self.freeze_dir,ESKY_CONTROL_DIR,"lockfile.txt"),os.path.join(self.freeze_dir,"esky-lockfile.txt"))
-        # TODO: remove compatability hook
-        shutil.copyfile(os.path.join(self.freeze_dir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt"),os.path.join(self.freeze_dir,"esky-bootstrap.txt"))
 
     def _run_create_zipfile(self):
         """Zip up the final distribution."""
@@ -320,7 +355,7 @@ class bdist_esky(Command):
             self.freezer_module.zipit(self,self.bootstrap_dir,zfname)
         else:
             create_zipfile(self.bootstrap_dir,zfname,compress=True)
-        shutil.rmtree(self.bootstrap_dir)
+        really_rmtree(self.bootstrap_dir)
 
     def _obj2code(self,obj):
         """Convert an object to some python source code.
@@ -496,7 +531,8 @@ class bdist_esky(Command):
         else:
             return ""
 
-    def get_msvcrt_private_assembly_files(self):
+    @staticmethod
+    def get_msvcrt_private_assembly_files():
         """Get (source,destination) tuples for the MSVCRT DLLs, manifest etc.
 
         This method generates data_files tuples for the MSVCRT DLLs, manifest
@@ -509,11 +545,13 @@ class bdist_esky(Command):
         it explicitly.  Before doing so, carefully check whether you have a
         license to distribute these files.
         """
-        msvcrt_info = self._get_msvcrt_info()
+        cls = bdist_esky
+        msvcrt_info = cls._get_msvcrt_info()
         if msvcrt_info is not None:
             msvcrt_name = msvcrt_info[0]
             #  Find installed manifest file with matching info
-            for manifest_file in self._find_msvcrt_manifest_files(msvcrt_name):
+            for candidate in cls._find_msvcrt_manifest_files(msvcrt_name):
+                manifest_file, msvcrt_dir = candidate
                 try:
                     with open(manifest_file,"rb") as mf:
                         manifest_data = mf.read()
@@ -528,25 +566,14 @@ class bdist_esky(Command):
                 err = "manifest for %s not found" % (msvcrt_info,)
                 raise RuntimeError(err)
             #  Copy the manifest and matching directory into the freeze dir.
-            #  The manifest file might be next to the dir, inside the dir, or
-            #  in a subdir named "Manifests".  Walk around till we find it.
-            msvcrt_dir = ".".join(manifest_file.split(".")[:-1])
-            if not os.path.isdir(msvcrt_dir):
-                msvcrt_basename = os.path.basename(msvcrt_dir)
-                msvcrt_parent = os.path.dirname(os.path.dirname(msvcrt_dir))
-                msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_basename)
-                if not os.path.isdir(msvcrt_dir):
-                    msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_name)
-                    if not os.path.isdir(msvcrt_dir):
-                        err = "manifest for %s not found" % (msvcrt_info,)
-                        raise RuntimeError(err)
             manifest_name = msvcrt_name + ".manifest"
             yield (manifest_file,os.path.join(msvcrt_name,manifest_name))
             for fnm in os.listdir(msvcrt_dir):
                 yield (os.path.join(msvcrt_dir,fnm),
                        os.path.join(msvcrt_name,fnm))
 
-    def _get_msvcrt_info(self):
+    @staticmethod
+    def _get_msvcrt_info():
         """Get info about the MSVCRT in use by this python executable.
 
         This parses the name, version and public key token out of the exe
@@ -565,8 +592,16 @@ class bdist_esky(Command):
                 return (name,version,pubkey)
         return None
         
-    def _find_msvcrt_manifest_files(self,name):
-        """Search the system for candidate MSVCRT manifest files."""
+    
+    @staticmethod
+    def _find_msvcrt_manifest_files(name):
+        """Search the system for candidate MSVCRT manifest files.
+
+        This method yields (manifest_file,msvcrt_dir) tuples giving a candidate
+        manifest file for the given assembly name, and the directory in which
+        the actual assembly data files are found.
+        """
+        cls = bdist_esky
         #  Search for redist files in a Visual Studio install
         progfiles = os.path.expandvars("%PROGRAMFILES%")
         for dnm in os.listdir(progfiles):
@@ -576,18 +611,94 @@ class bdist_esky(Command):
                     for fnm in filenames:
                         if name.lower() in fnm.lower():
                             if fnm.lower().endswith(".manifest"):
-                                yield os.path.join(subdir,fnm)
+                                mf = os.path.join(subdir,fnm)
+                                md = cls._find_msvcrt_dir_for_manifest(name,mf)
+                                if md is not None:
+                                    yield (mf,md)
         #  Search for manifests installed in the WinSxS directory
         winsxs_m = os.path.expandvars("%WINDIR%\\WinSxS\\Manifests")
         for fnm in os.listdir(winsxs_m):
             if name.lower() in fnm.lower():
                 if fnm.lower().endswith(".manifest"):
-                    yield os.path.join(winsxs_m,fnm)
+                    mf = os.path.join(winsxs_m,fnm)
+                    md = cls._find_msvcrt_dir_for_manifest(name,mf)
+                    if md is not None:
+                        yield (mf,md)
         winsxs = os.path.expandvars("%WINDIR%\\WinSxS")
         for fnm in os.listdir(winsxs):
             if name.lower() in fnm.lower():
                 if fnm.lower().endswith(".manifest"):
-                    yield os.path.join(winsxs,fnm)
+                    mf = os.path.join(winsxs,fnm)
+                    md = cls._find_msvcrt_dir_for_manifest(name,mf)
+                    if md is not None:
+                        yield (mf,md)
+
+    @staticmethod
+    def _find_msvcrt_dir_for_manifest(msvcrt_name,manifest_file):
+        """Find the directory containing data files for the given manifest.
+
+        This searches a few common locations for the data files that go with
+        the given manifest file.  If a suitable directory is found then it is
+        returned, otherwise None is returned.
+        """
+        #  The manifest file might be next to the dir, inside the dir, or
+        #  in a subdir named "Manifests".  Walk around till we find it.
+        msvcrt_dir = ".".join(manifest_file.split(".")[:-1])
+        if os.path.isdir(msvcrt_dir):
+            return msvcrt_dir
+        msvcrt_basename = os.path.basename(msvcrt_dir)
+        msvcrt_parent = os.path.dirname(os.path.dirname(msvcrt_dir))
+        msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_basename)
+        if os.path.isdir(msvcrt_dir):
+            return msvcrt_dir
+        msvcrt_dir = os.path.join(msvcrt_parent,msvcrt_name)
+        if os.path.isdir(msvcrt_dir):
+            return msvcrt_dir
+        return None
+
+
+    def compile_to_bootstrap_exe(self,exe,source):
+        """Compile the given sourcecode into a bootstrapping exe.
+
+        This method compiles the given sourcecode into a stand-alone exe using
+        PyPy, then stores that in the bootstrap env under the name of the given
+        Executable object.  If the source has been previously compiled then a
+        cached version of the exe may be used.
+        """
+        source = "__rpython__ = True\n" + source
+        cdir = os.path.join(self.tempdir,"compile")
+        if not os.path.exists(cdir):
+            os.mkdir(cdir)
+        source_hash = hashlib.md5(source).hexdigest()
+        outname = "bootstrap_%s.%s" % (source_hash,get_platform())
+        if exe.gui_only:
+            outname += ".gui"
+        if sys.platform == "win32":
+            outname += ".exe"
+        #  First try to use a precompiled version.
+        if COMPILED_BOOTSTRAP_CACHE is not None:
+            outfile = os.path.join(COMPILED_BOOTSTRAP_CACHE,outname)
+            if os.path.exists(outfile):
+                return self.copy_to_bootstrap_env(outfile,exe.name)
+        #  Otherwise we have to compile it anew.
+        try:
+            outfile = self._compiled_exes[(source_hash,exe.gui_only)]
+        except KeyError:
+            infile = os.path.join(cdir,"bootstrap.py")
+            outfile = os.path.join(cdir,outname)
+            with open(infile,"wt") as f:
+                f.write(source)
+            opts = dict(gui_only=exe.gui_only)
+            pypyc.compile_rpython(infile,outfile,**opts)
+            self._compiled_exes[(source_hash,exe.gui_only)] = outfile
+        #  Try to save the compiled exe for future use.
+        if COMPILED_BOOTSTRAP_CACHE is not None:
+            cachedfile = os.path.join(COMPILED_BOOTSTRAP_CACHE,outname)
+            try:
+                shutil.copy2(outfile,cachedfile)
+            except EnvironmentError:
+                pass
+        return self.copy_to_bootstrap_env(outfile,exe.name)
 
     def copy_to_bootstrap_env(self,src,dst=None):
         """Copy the named file into the bootstrap environment.
@@ -690,3 +801,5 @@ distutils.command.__all__.append("bdist_esky_patch")
 sys.modules["distutils.command.bdist_esky"] = sys.modules["esky.bdist_esky"]
 sys.modules["distutils.command.bdist_esky_patch"] = sys.modules["esky.bdist_esky"]
 
+
+
