@@ -38,6 +38,8 @@ def freeze(dist):
         includes.append(inc)
     for exc in options.pop("excludes",()):
         excludes.append(exc)
+    if "pypy" not in includes and "pypy" not in excludes:
+        excludes.append("pypy")
     options["includes"] = includes
     options["excludes"] = excludes
     # py2app can't simultaneously freeze multiple scripts.
@@ -86,52 +88,62 @@ def freeze(dist):
     for (src,arcnm) in dist.get_package_data():
         lib.write(src,arcnm)
     lib.close()
-    #  Copy the core dependencies into the bootstrap env.
-    pydir = "python%d.%d" % sys.version_info[:2]
-    def copy_to_bootstrap_env(src,dst=None):
-        if dst is None:
-            dst = src
-        src = os.path.join(appnm,src)
-        dist.copy_to_bootstrap_env(src,dst)
+    #  Create the bootstraping code, using custom code if specified.
+    code_source = ["__esky_name__ = '%s'" % (dist.distribution.get_name(),)]
+    code_source.append(inspect.getsource(esky.bootstrap))
+    if not dist.compile_bootstrap_exes:
+        code_source.append(_FAKE_ESKY_BOOTSTRAP_MODULE)
+        code_source.append(_EXTRA_BOOTSTRAP_CODE)
+    code_source.append(dist.get_bootstrap_code())
+    code_source.append("if not __rpython__:")
+    code_source.append("    bootstrap()")
+    code_source = "\n".join(code_source)
+    if dist.compile_bootstrap_exes:
+        for exe in dist.get_executables(normalise=False):
+            if not exe.include_in_bootstrap_env:
+                continue
+            dist.compile_to_bootstrap_exe(exe,code_source)
+    else:
+        #  Copy the core dependencies into the bootstrap env.
+        pydir = "python%d.%d" % sys.version_info[:2]
+        def copy_to_bootstrap_env(src,dst=None):
+            if dst is None:
+                dst = src
+            src = os.path.join(appnm,src)
+            dist.copy_to_bootstrap_env(src,dst)
+        copy_to_bootstrap_env("Contents/Frameworks/Python.framework")
+        copy_to_bootstrap_env("Contents/Resources/include")
+        copy_to_bootstrap_env("Contents/Resources/lib/"+pydir+"/config")
+        if "fcntl" not in sys.builtin_module_names:
+            dynload = "Contents/Resources/lib/"+pydir+"/lib-dynload"
+            for nm in os.listdir(os.path.join(app_dir,dynload)):
+                if nm.startswith("fcntl"):
+                    copy_to_bootstrap_env(os.path.join(dynload,nm))
+        copy_to_bootstrap_env("Contents/Resources/__error__.sh")
+        copy_to_bootstrap_env("Contents/Resources/__boot__.py")
+        copy_to_bootstrap_env("Contents/Resources/site.py")
+        #  Copy the bootstrapping code into the __boot__.py file.
+        bsdir = dist.boostrap_dir
+        with open(bsdir+"/Contents/Resources/__boot__.py","wt") as f:
+            f.write(code_source)
+        #  Clear site.py in the bootstrap dir, it doesn't do anything useful.
+        with open(bsdir+"/Contents/Resources/site.py","wt") as f:
+            f.write("")
+        #  Copy the loader program for each script into the bootstrap env.
+        copy_to_bootstrap_env("Contents/MacOS/python")
+        for exe in dist.get_executables(normalise=False):
+            if not exe.include_in_bootstrap_env:
+                continue
+            exepath = copy_to_bootstrap_env("Contents/MacOS/"+exe.name)
+    #  Copy non-python resources (e.g. icons etc) into the bootstrap dir
     copy_to_bootstrap_env("Contents/Info.plist")
     copy_to_bootstrap_env("Contents/PkgInfo")
-    copy_to_bootstrap_env("Contents/Frameworks/Python.framework")
-    copy_to_bootstrap_env("Contents/Resources/include")
-    copy_to_bootstrap_env("Contents/Resources/lib/"+pydir+"/config")
-    if "fcntl" not in sys.builtin_module_names:
-        dynload = "Contents/Resources/lib/"+pydir+"/lib-dynload"
-        for nm in os.listdir(os.path.join(app_dir,dynload)):
-            if nm.startswith("fcntl"):
-                copy_to_bootstrap_env(os.path.join(dynload,nm))
-    copy_to_bootstrap_env("Contents/Resources/__error__.sh")
-    copy_to_bootstrap_env("Contents/Resources/__boot__.py")
-    copy_to_bootstrap_env("Contents/Resources/site.py")
-    #  Copy any other mentioned files (icons etc) into the bootstrap env
     with open(os.path.join(app_dir,"Contents","Info.plist"),"rt") as f:
         infotxt = f .read()
     for nm in os.listdir(os.path.join(app_dir,"Contents","Resources")):
         if nm in infotxt:
             copy_to_bootstrap_env("Contents/Resources/"+nm)
-    #  Create the bootstraping code, using custom code if specified.
-    #  It gets stored as plain python code in Contents/Resources/__boot__.py
-    code_source = [inspect.getsource(esky.bootstrap)]
-    code_source.append(_FAKE_ESKY_BOOTSTRAP_MODULE)
-    code_source.append(_EXTRA_BOOTSTRAP_CODE)
-    code_source.append("__esky_name__ = '%s'" % (dist.distribution.get_name(),))
-    code_source.append(dist.get_bootstrap_code())
-    code_source.append("bootstrap()")
-    code_source = "\n".join(code_source)
-    with open(dist.bootstrap_dir+"/Contents/Resources/__boot__.py","wt") as f:
-        f.write("".join(code_source))
-    #  Clear site.py in the bootstrap dir, it doesn't do anything useful.
-    with open(dist.bootstrap_dir+"/Contents/Resources/site.py","wt") as f:
-        f.write("")
-    #  Copy the loader program for each script into the bootstrap env.
-    copy_to_bootstrap_env("Contents/MacOS/python")
-    for exe in dist.get_executables(normalise=False):
-        if not exe.include_in_bootstrap_env:
-            continue
-        exepath = copy_to_bootstrap_env("Contents/MacOS/"+exe.name)
+
 
 
 def zipit(dist,bsdir,zfname):
