@@ -15,6 +15,10 @@ of bootstrapping into apps made with older versions of esky, since a partial
 update could result in the boostrapper from a new version being forced
 to load an old version.
 
+If you want to compile your bootstrapping exes into standalone executables,
+this module must also be written in the "RPython" dialect used by the PyPy
+translation toolchain.
+
 The code from this module is always executed in the bootstrapping environment
 before any custom bootstrapping code.  It provides the following functions for
 use during the bootstrap process:
@@ -29,22 +33,61 @@ use during the bootstrap process:
 
 """
 
+
 import sys
 import errno
 
-ESKY_CONTROL_DIR = "esky-files"
+try:
+    ESKY_CONTROL_DIR
+except NameError:
+    ESKY_CONTROL_DIR = "esky-files"
+
+try:
+    ESKY_APPDATA_DIR
+except NameError:
+    ESKY_APPDATA_DIR = "appdata"
+
+try:
+    __esky_name__
+except NameError:
+    __esky_name__ = ""
+
+try:
+    __rpython__
+except NameError:
+    __rpython__ = False
+
+#  RPython doesn't handle SystemExit automatically, so we put the exit code
+#  in this global var and catch SystemExit ourselves at the outmost scope.
+_exit_code = [0]
 
 #  The os module is not builtin, so we grab what we can from the
 #  platform-specific modules and fudge the rest.
 if "posix" in sys.builtin_module_names:
     import fcntl
-    from posix import listdir, stat, unlink, rename, execv
+    from posix import listdir, stat, unlink, rename, execv, getcwd, environ
+    from posix import open as os_open
+    from posix import read as os_read
+    from posix import close as os_close
     SEP = "/"
+    def isabs(path):
+        return (path.startswith(SEP))
 elif "nt" in sys.builtin_module_names:
     fcntl = None
     import nt
-    from nt import listdir, stat, unlink, rename, spawnv, P_WAIT
+    from nt import listdir, stat, unlink, rename, spawnv
+    from nt import getcwd, P_WAIT, environ
+    from nt import open as os_open
+    from nt import read as os_read
+    from nt import close as os_close
     SEP = "\\"
+    def isabs(path):
+        if path.startswith(SEP):
+            return True
+        if len(path) >= 2:
+            if path[0].isalpha() and path[1] == ":":
+                return True
+        return False
     #  The standard execv terminates the spawning process, which makes
     #  it impossible to wait for it.  This alternative is waitable, and
     #  uses the esky.slaveproc machinery to avoid leaving zombie children.
@@ -52,11 +95,13 @@ elif "nt" in sys.builtin_module_names:
         #  Create an O_TEMPORARY file and pass its name to the slave process.
         #  When this master process dies, the file will be deleted and the
         #  slave process will know to terminate.
-        tdir = nt.environ.get("TEMP",None)
+        try:
+            tdir = environ["TEMP"]
+        except:
+            tdir = None
         if tdir:
-            tfile = None
             try:
-                nt.mkdir(pathjoin(tdir,"esky-slave-procs"))
+                nt.mkdir(pathjoin(tdir,"esky-slave-procs"),0600)
             except EnvironmentError:
                 pass
             if exists(pathjoin(tdir,"esky-slave-procs")):
@@ -65,26 +110,137 @@ elif "nt" in sys.builtin_module_names:
                     tfilenm = "slave-%d.%d.txt" % (nt.getpid(),i,)
                     tfilenm = pathjoin(tdir,"esky-slave-procs",tfilenm)
                     try:
-                        tfile = nt.open(tfilenm,flags)
+                        os_open(tfilenm,flags,0600)
+                        args.insert(1,tfilenm)
+                        args.insert(1,"--esky-slave-proc")
                         break
                     except EnvironmentError:
-                        raise
                         pass
-        if tdir and tfile:
-            args.insert(1,tfilenm)
-            args.insert(1,"--esky-slave-proc")
         res = spawnv(P_WAIT,filename,args)
+        _exit_code[0] = res
         raise SystemExit(res)
+    #  A fake fcntl module which is false, but can fake out RPython
+    class fcntl:
+        LOCK_SH = 0
+        def flock(self,fd,mode):
+            pass
+        def __nonzero__(self):
+            return False
+    fcntl = fcntl()
 else:
     raise RuntimeError("unsupported platform: " + sys.platform)
 
 
-__esky_name__ = None
+if __rpython__:
+    # RPython provides ll hooks for the actual os.environ object, not the
+    # one we pulled out of "nt" or "posix".
+    from os import environ
+
+    # RPython doesn't have access to the "sys" module, so we fake it out.
+    # The entry_point function will set these value appropriately.
+    _sys = sys
+    class sys:
+        platform = _sys.platform
+        executable = _sys.executable
+        argv = _sys.argv
+        version_info = _sys.version_info
+        modules = {}
+        builtin_module_names = _sys.builtin_module_names
+        def exit(self,code):
+            _exit_code[0] = code
+            raise SystemExit(code)
+        def exc_info(self):
+            return None,None,None
+    sys = sys()
+    sys.modules["sys"] = sys
+
+    #  RPython doesn't provide the sorted() builtin, and actually makes sorting
+    #  quite complicated in general.  I can't convince the type annotator to be
+    #  happy about using their "listsort" module, so I'm doing my own using a
+    #  simple insertion sort.  We're only sorting short lists and they always
+    #  contain (list(str),str), so this should do for now.
+    def _list_gt(l1,l2):
+        i = 0
+        while i < len(l1) and i < len(l2):
+            if l1[i] > l2[i]:
+               return True
+            if l1[i] < l2[i]:
+               return False
+            i += 1
+        if len(l1) > len(l2):
+            return True
+        return False
+    def sorted(lst,reverse=False):
+        slst = []
+        if reverse:
+            for item in lst:
+                for j in xrange(len(slst)):
+                    if not _list_gt(slst[j][0],item[0]):
+                        slst.insert(j,item)
+                        break 
+                else:
+                    slst.append(item)
+        else:
+            for item in lst:
+                for j in xrange(len(slst)):
+                    if _list_gt(slst[j][0],item[0]):
+                        slst.insert(j,item)
+                        break 
+                else:
+                    slst.append(item)
+        return slst
+    # RPython doesn't provide the "zfill" or "isalnum" methods on strings.
+    def zfill(str,n):
+        while len(str) < n:
+            str = "0" + str
+        return str
+    def isalnum(str):
+        for c in str:
+            if not c.isalnum():
+                return False
+        return True
+    # RPython doesn't provide the "fcntl" module.  Fake it.
+    # TODO: implement it using externals
+    if fcntl:
+        class fcntl:
+            LOCK_SH = fcntl.LOCK_SH
+            def flock(self,fd,mode):
+                pass
+        fcntl = fcntl()
+else:
+    #  We need to use a compatability wrapper for some string methods missing
+    #  in RPython, since we can't just add them as methods on the str type.
+    def zfill(str,n):
+        return str.zfill(n)
+    def isalnum(str):
+        return str.isalnum()
 
 
 def pathjoin(*args):
     """Local re-implementation of os.path.join."""
-    return SEP.join(args)
+    path = args[0]
+    for arg in list(args[1:]):
+        if isabs(arg):
+            path = arg
+        else:
+            while path.endswith(SEP):
+                path = path[:-1]
+            path = path + SEP + arg
+    return path
+
+def abspath(path):
+    """Absolute-ize and normalise the given path."""
+    path = pathjoin(getcwd(),path)
+    components_in = path.split(SEP)
+    components = [components_in[0]]
+    for comp in components_in[1:]:
+        if not comp or comp == ".":
+            pass
+        elif comp == "..":
+            components.pop()
+        else:
+            components.append(comp)
+    return SEP.join(components)
 
 def basename(p):
     """Local re-implementation of os.path.basename."""
@@ -98,12 +254,12 @@ def exists(path):
     """Local re-implementation of os.path.exists."""
     try:
         stat(path)
-    except EnvironmentError:
-        e = sys.exc_info()[1]
-        if e.errno not in (errno.ENOENT,errno.ENOTDIR,errno.ESRCH,):
-            raise
-        else:
-            return False
+    except EnvironmentError, e:
+        # TODO: how to get the errno under RPython?
+        if not __rpython__:
+            if e.errno not in (errno.ENOENT,errno.ENOTDIR,errno.ESRCH,):
+                raise
+        return False
     else:
         return True
 
@@ -135,14 +291,30 @@ def bootstrap():
     chain-loads that version of the application.
     """
     appdir = appdir_from_executable(sys.executable)
+    vsdir = pathjoin(appdir,ESKY_APPDATA_DIR)
+    # TODO: remove compatability hook for ESKY_APPDATA_DIR="".
     best_version = None
-    if __esky_name__ is not None:
-        best_version = get_best_version(appdir,appname=__esky_name__)
-    if best_version is None:
-        best_version = get_best_version(appdir)
-    if best_version is None:
-        raise RuntimeError("no usable frozen versions were found")
-    return chainload(pathjoin(appdir,best_version))
+    try:
+        if __esky_name__:
+            best_version = get_best_version(vsdir,appname=__esky_name__)
+        if best_version is None:
+            best_version = get_best_version(vsdir)
+        if best_version is None:
+            if exists(vsdir):
+                raise RuntimeError("no usable frozen versions were found")
+            else:
+                raise EnvironmentError
+    except EnvironmentError:
+        if exists(vsdir):
+            raise
+        vsdir = appdir
+        if __esky_name__:
+            best_version = get_best_version(vsdir,appname=__esky_name__)
+        if best_version is None:
+            best_version = get_best_version(vsdir)
+        if best_version is None:
+            raise RuntimeError("no usable frozen versions were found")
+    return chainload(pathjoin(vsdir,best_version))
 
 
 def chainload(target_dir):
@@ -161,10 +333,11 @@ def chainload(target_dir):
         #  If the bootstrap file is missing, the version is being uninstalled.
         #  Our only option is to re-execute ourself and find the new version.
         if exists(dirname(target_dir)):
-            if not exists(pathjoin(target_dir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")):
-                # TODO: remove compatability hook
-                if not exists(pathjoin(target_dir,"esky-bootstrap.txt")):
-                    execv(sys.executable,sys.argv)
+            bsfile = pathjoin(target_dir,ESKY_CONTROL_DIR)
+            bsfile = pathjoin(bsfile,"bootstrap-manifest.txt")
+            if not exists(bsfile):
+                execv(sys.executable,list(sys.argv))
+                return
         raise
     else:
         #  If all goes well, we can actually launch the target version.
@@ -173,27 +346,39 @@ def chainload(target_dir):
 
 def get_exe_locations(target_dir):
     """Generate possible locations from which to chainload in the target dir."""
+    # TODO: let this be a generator when not compiling with PyPy, so we can
+    # avoid a few stat() calls in the common case.
+    locs = []
     appdir = appdir_from_executable(sys.executable)
     #  If we're in an appdir, first try to launch from within "<appname>.app"
     #  directory.  We must also try the default scheme for backwards compat.
     if sys.platform == "darwin":
         if basename(dirname(sys.executable)) == "MacOS":
-            if __esky_name__ is None:
-                yield pathjoin(target_dir,
-                               __esky_name__+".app",
-                               sys.executable[len(appdir)+1:])
+            if __esky_name__:
+                locs.append(pathjoin(target_dir,
+                                     __esky_name__+".app",
+                                     sys.executable[len(appdir)+1:]))
             else:
                 for nm in listdir(target_dir):
                     if nm.endswith(".app"):
-                        yield pathjoin(target_dir,
-                                       nm,
-                                       sys.executable[len(appdir)+1:])
+                        locs.append(pathjoin(target_dir,
+                                             nm,
+                                             sys.executable[len(appdir)+1:]))
     #  This is the default scheme: the same path as the exe in the appdir.
-    yield target_dir + sys.executable[len(appdir):]
+    locs.append(target_dir + sys.executable[len(appdir):])
     #  If sys.executable was a backup file, try using original filename.
     orig_exe = get_original_filename(sys.executable)
     if orig_exe is not None:
-        yield target_dir + orig_executable[len(appdir):]
+        locs.append(target_dir + orig_exe[len(appdir):])
+    return locs
+
+
+def verify(target_file):
+    """Verify the integrity of the given target file.
+
+    By default this is a no-op; override it to provide e.g. signature checks.
+    """
+    pass
 
 
 def _chainload(target_dir):
@@ -204,15 +389,25 @@ def _chainload(target_dir):
     """
     exc_type,exc_value,traceback = None,None,None
     for target_exe in get_exe_locations(target_dir):
+        verify(target_exe)
         try:
             execv(target_exe,[target_exe] + sys.argv[1:])
-        except EnvironmentError:
-            exc_type,exc_value,traceback = sys.exc_info()
-            if exc_value.errno != errno.ENOENT:
-                raise
+            return
+        except EnvironmentError, exc_value:
+            #  Careful, RPython lacks a usable exc_info() function.
+            exc_type,_,traceback = sys.exc_info()
+            if not __rpython__:
+                if exc_value.errno != errno.ENOENT:
+                    raise
+            else:
+                if exists(target_exe):
+                    raise
     else:
         if exc_value is not None:
-            raise exc_type,exc_value,traceback
+            if exc_type is not None:
+                raise exc_type,exc_value,traceback
+            else:
+                raise exc_value
 
 
 def get_best_version(appdir,include_partial_installs=False,appname=None):
@@ -286,9 +481,6 @@ def is_version_dir(vdir):
     """
     if exists(pathjoin(vdir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")):
         return True
-    # TODO: remove compatability hook
-    if exists(pathjoin(vdir,"esky-bootstrap.txt")):
-        return True
     return False
 
 
@@ -298,10 +490,8 @@ def is_installed_version_dir(vdir):
     Currently, a completed installation is indicated by the lack of an
     "esky-files/bootstrap" directory.
     """
-    # TODO: remove compatability hook
     if not exists(pathjoin(vdir,ESKY_CONTROL_DIR,"bootstrap")):
-        if not exists(pathjoin(vdir,"esky-bootstrap")):
-            return True
+        return True
     return False
 
 
@@ -313,9 +503,6 @@ def is_uninstalled_version_dir(vdir):
     """
     if exists(pathjoin(vdir,ESKY_CONTROL_DIR,"bootstrap-manifest-old.txt")):
         return True
-    # TODO: remove compatability hooks
-    if exists(pathjoin(vdir,"esky-bootstrap-old.txt")):
-        return True
     return False
     
 
@@ -329,7 +516,7 @@ def split_app_version(s):
     idx = 1
     while idx < len(bits):
         if bits[idx]:
-            if not bits[idx][0].isalpha() or not bits[idx].isalnum():
+            if not bits[idx][0].isalpha() or not isalnum(bits[idx]):
                 break
         idx += 1
     appname = "-".join(bits[:idx])
@@ -350,12 +537,12 @@ def join_app_version(appname,version,platform):
 def parse_version(s):
     """Parse a version string into a chronologically-sortable key
 
-    This function returns a tuple of strings that compares with the results
+    This function returns a sequence of strings that compares with the results
     for other versions in a chronologically sensible way.  You'd use it to
     compare two version strings like so:
 
         if parse_version("1.9.2") > parse_version("1.10.0"):
-            print "what rubbish, that's an older version!"
+            raise RuntimeError("what rubbish, that's an older version!")
 
     This is essentially the parse_version() function from pkg_resources,
     but re-implemented to avoid using modules that may not be available
@@ -370,20 +557,22 @@ def parse_version(s):
             while parts and parts[-1]=='00000000':
                 parts.pop()
         parts.append(part)
-    return tuple(parts)
+    return parts
 
 
 _replace_p = {'pre':'c', 'preview':'c','-':'final-','rc':'c','dev':'@'}.get
 def _parse_version_parts(s):
+    parts = []
     for part in _split_version_components(s):
         part = _replace_p(part,part)
         if not part or part=='.':
             continue
         if part[:1] in '0123456789':
-            yield part.zfill(8)    # pad for numeric comparison
+            parts.append(zfill(part,8))    # pad for numeric comparison
         else:
-            yield '*'+part
-    yield '*final'  # ensure that alpha/beta/candidate are before final
+            parts.append('*'+part)
+    parts.append('*final')  # ensure that alpha/beta/candidate are before final
+    return parts
 
 
 def _split_version_components(s):
@@ -393,6 +582,7 @@ def _split_version_components(s):
     Unfortunately the 're' module isn't in the bootstrap, so we have to do
     an equivalent parse by hand.  Forunately, that's pretty easy.
     """
+    comps = []
     start = 0
     while start < len(s):
         end = start+1
@@ -407,8 +597,9 @@ def _split_version_components(s):
         else:
             while end < len(s) and not (s[end].isdigit() or s[end].isalpha() or s[end] in (".","-")):
                 end += 1
-        yield s[start:end]
+        comps.append(s[start:end])
         start = end
+    return comps
 
 
 def get_original_filename(backname):
@@ -420,11 +611,11 @@ def get_original_filename(backname):
 
     If no matching original file is found, None is returned.
     """
-    filtered = ".".join(filter(lambda n: n != "old",backname.split(".")))
+    filtered = ".".join([n for n in backname.split(".") if n != "old"])
     for nm in listdir(dirname(backname)):
         if nm == backname:
             continue
-        if filtered == ".".join(filter(lambda n: n != "old",nm.split("."))):
+        if filtered == ".".join([n for n in nm.split(".") if n != "old"]):
             return pathjoin(dirname(backname),nm)
     return None
 
@@ -437,10 +628,7 @@ def lock_version_dir(vdir):
         #  On win32, we just hold bootstrap file open for reading.
         #  This will prevent it from being renamed during uninstall.
         lockfile = pathjoin(vdir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")
-        # TODO: remove compatability hooks
-        if not exists(lockfile):
-            lockfile = pathjoin(vdir,"esky-bootstrap.txt")
-        _locked_version_dirs.setdefault(vdir,[]).append(open(lockfile,"rt"))
+        _locked_version_dirs.setdefault(vdir,[]).append(os_open(lockfile,0,0))
     else:
         #  On posix platforms we take a shared flock on esky-files/lockfile.txt.
         #  While fcntl.fcntl locks are apparently the new hotness, they have
@@ -448,17 +636,31 @@ def lock_version_dir(vdir):
         #      * not inherited across fork()
         #      * released when closing *any* fd associated with that file
         #  fcntl.flock doesn't have these problems, but may fail on NFS.
-        #  To complicate matters, python sometimes emulated flock with fcntl!
+        #  To complicate matters, python sometimes emulates flock with fcntl!
         #  We therefore use a separate lock file to avoid unpleasantness.
         lockfile = pathjoin(vdir,ESKY_CONTROL_DIR,"lockfile.txt")
-        # TODO: remove compatability hooks
-        if not exists(lockfile):
-            lockfile = pathjoin(vdir,"esky-lockfile.txt")
-        f = open(lockfile,"r")
+        f = os_open(lockfile,0,0)
         _locked_version_dirs.setdefault(vdir,[]).append(f)
         fcntl.flock(f,fcntl.LOCK_SH)
 
 def unlock_version_dir(vdir):
     """Unlock the given version dir, allowing it to be uninstalled."""
-    _locked_version_dirs[vdir].pop().close()
+    os_close(_locked_version_dirs[vdir].pop())
+
+if __rpython__:
+    def main():
+        bootstrap()
+    def target(driver,args):
+        """Target function for compiling a standalone bootstraper with PyPy."""
+        def entry_point(argv):
+             exit_code = 0
+             #  Todo: resolve symlinks etc
+             sys.executable = abspath(pathjoin(getcwd(),argv[0]))
+             sys.argv = argv
+             try:
+                 main()
+             except SystemExit, e:
+                 exit_code = _exit_code[0]
+             return exit_code
+        return entry_point, None
 
