@@ -61,6 +61,7 @@ except NameError:
 import os
 import sys
 import bz2
+import time
 import shutil
 import hashlib
 import optparse
@@ -94,12 +95,13 @@ PATCH_HEADER = "ESKYPTCH".encode("ascii")
 
 from esky.errors import Error
 from esky.util import extract_zipfile, create_zipfile, deep_extract_zipfile,\
-                      zipfile_common_prefix_dir
+                      zipfile_common_prefix_dir, really_rmtree, really_rename
 
 __all__ = ["PatchError","DiffError","main","write_patch","apply_patch",
            "Differ","Patcher"]
 
 
+
 class PatchError(Error):
     """Error raised when a patch fails to apply."""
     pass
@@ -301,7 +303,7 @@ class Patcher(object):
         if self.outfile:
             self.outfile.close()
         if self._workdir and shutil:
-            shutil.rmtree(self._workdir)
+            really_rmtree(self._workdir)
 
     def _read(self,size):
         """Read the given number of bytes from the command stream."""
@@ -350,7 +352,7 @@ class Patcher(object):
         """
         if not self.outfile and not self.dry_run:
             if os.path.exists(self.target) and not os.path.isfile(self.target):
-                shutil.rmtree(self.target)
+                really_rmtree(self.target)
             self.new_target = self.target + ".new"
             while os.path.exists(self.new_target):
                 self.new_target += ".new"
@@ -372,8 +374,10 @@ class Patcher(object):
             self.outfile.close()
             self.outfile = None
             if os.path.exists(self.target):
-               os.unlink(self.target)
-            os.rename(self.new_target,self.target)
+                os.unlink(self.target)
+                if sys.platform == "win32":
+                    time.sleep(0.01)
+            really_rename(self.new_target,self.target)
             self.new_target = None
 
     def _check_path(self,path=None):
@@ -422,6 +426,13 @@ class Patcher(object):
                 getattr(self,"_do_" + _COMMANDS[cmd])()
         except EOFError:
             self._check_end_patch()
+        finally:
+            if self.infile:
+                self.infile.close()
+                self.infile = None
+            if self.outfile:
+                self.outfile.close()
+                self.outfile = None
 
     def _do_END(self):
         """Execute the END command.
@@ -505,7 +516,7 @@ class Patcher(object):
         self._check_end_patch()
         if not self.dry_run:
             if os.path.isdir(self.target):
-                shutil.rmtree(self.target)
+                really_rmtree(self.target)
             elif os.path.exists(self.target):
                 os.unlink(self.target)
             os.makedirs(self.target)
@@ -518,7 +529,7 @@ class Patcher(object):
         self._check_end_patch()
         if not self.dry_run:
             if os.path.isdir(self.target):
-                shutil.rmtree(self.target)
+                really_rmtree(self.target)
             elif os.path.exists(self.target):
                 os.unlink(self.target)
 
@@ -537,7 +548,7 @@ class Patcher(object):
         if not self.dry_run:
             if os.path.exists(self.target):
                 if os.path.isdir(self.target):
-                    shutil.rmtree(self.target)
+                    really_rmtree(self.target)
                 else:
                     os.unlink(self.target)
             if os.path.isfile(source_path):
@@ -560,10 +571,12 @@ class Patcher(object):
         if not self.dry_run:
             if os.path.exists(self.target):
                 if os.path.isdir(self.target):
-                    shutil.rmtree(self.target)
+                    really_rmtree(self.target)
                 else:
                     os.unlink(self.target)
-            os.rename(source_path,self.target)
+                if sys.platform == "win32":
+                    time.sleep(0.01)
+            really_rename(source_path,self.target)
 
     def _do_PF_COPY(self):
         """Execute the PF_COPY command.
@@ -649,7 +662,7 @@ class Patcher(object):
             m_temp = os.path.join(workdir,"meta")
             z_temp = os.path.join(workdir,"result.zip")
         cur_state = self._blank_state()
-        zfmeta = [None]
+        zfmeta = [None]  # stupid lack of mutable closure variables...
         #  First we process a set of commands to generate the zipfile metadata.
         def end_metadata():
             if not self.dry_run:
@@ -666,7 +679,7 @@ class Patcher(object):
                         self.outfile.write(data)
                         data = f.read(1024*16)
                 zfmeta[0].close()
-                shutil.rmtree(workdir)
+                really_rmtree(workdir)
         self._context_stack.append(end_contents)
         self._context_stack.append(end_metadata)
         if not self.dry_run:
@@ -941,48 +954,47 @@ class Differ(object):
         bsdiff.
         """
         spos = 0
-        tfile = open(target,"rb")
-        if os.path.isfile(source):
-            sfile = open(source,"rb")
-        else:
-            sfile = None
-        try:
-            #  Process the file in diff_window_size blocks.  This
-            #  will produce slightly bigger patches but we avoid
-            #  running out of memory for large files.
-            tdata = tfile.read(self.diff_window_size)
-            if not tdata:
-                #  The file is empty, do a raw insert of zero bytes.
-                self._write_command(PF_INS_RAW)
-                self._write_bytes("".encode("ascii"))
+        with open(target,"rb") as tfile:
+            if os.path.isfile(source):
+                sfile = open(source,"rb")
             else:
-                while tdata:
-                    sdata = ""
-                    if sfile is not None:
-                        sdata = sfile.read(self.diff_window_size)
-                    #  Look for a shared prefix.
-                    i = 0; maxi = min(len(tdata),len(sdata))
-                    while i < maxi and tdata[i] == sdata[i]:
-                        i += 1
-                    #  Copy it in directly, unless it's tiny.
-                    if i > 8:
-                        skipbytes = sfile.tell() - len(sdata) - spos
-                        if skipbytes > 0:
-                            self._write_command(PF_SKIP)
-                            self._write_int(skipbytes)
-                            spos += skipbytes
-                        self._write_command(PF_COPY)
-                        self._write_int(i)
-                        tdata = tdata[i:]; sdata = sdata[i:]
-                        spos += i
-                    #  Write the rest of the block as a diff
-                    if tdata:
-                        spos += self._write_file_patch(sdata,tdata)
-                    tdata = tfile.read(self.diff_window_size)
-        finally:
-            tfile.close()
-            if sfile:
-                sfile.close()
+                sfile = None
+            try:
+                #  Process the file in diff_window_size blocks.  This
+                #  will produce slightly bigger patches but we avoid
+                #  running out of memory for large files.
+                tdata = tfile.read(self.diff_window_size)
+                if not tdata:
+                    #  The file is empty, do a raw insert of zero bytes.
+                    self._write_command(PF_INS_RAW)
+                    self._write_bytes("".encode("ascii"))
+                else:
+                    while tdata:
+                        sdata = ""
+                        if sfile is not None:
+                            sdata = sfile.read(self.diff_window_size)
+                        #  Look for a shared prefix.
+                        i = 0; maxi = min(len(tdata),len(sdata))
+                        while i < maxi and tdata[i] == sdata[i]:
+                            i += 1
+                        #  Copy it in directly, unless it's tiny.
+                        if i > 8:
+                            skipbytes = sfile.tell() - len(sdata) - spos
+                            if skipbytes > 0:
+                                self._write_command(PF_SKIP)
+                                self._write_int(skipbytes)
+                                spos += skipbytes
+                            self._write_command(PF_COPY)
+                            self._write_int(i)
+                            tdata = tdata[i:]; sdata = sdata[i:]
+                            spos += i
+                        #  Write the rest of the block as a diff
+                        if tdata:
+                            spos += self._write_file_patch(sdata,tdata)
+                        tdata = tfile.read(self.diff_window_size)
+            finally:
+                if sfile is not None:
+                    sfile.close()
 
     def _find_similar_sibling(self,source,target,nm):
         """Find a sibling of an entry against which we can calculate a diff.
@@ -1055,7 +1067,7 @@ class _tempdir(object):
     def __enter__(self):
         return self.path
     def __exit__(self,*args):
-        shutil.rmtree(self.path)
+        really_rmtree(self.path)
 
 
 if cx_bsdiff is not None:
@@ -1209,6 +1221,7 @@ def main(args):
             scale = 1024 * 1024 * 1024
             opts.diff_window = opts.diff_window[:-1]
         opts.diff_window = int(float(opts.diff_window)*scale)
+    stream = None
     try:
         cmd = args[0]
         if cmd == "diff":
@@ -1270,12 +1283,16 @@ def main(args):
                     create_zipfile(target,target_temp)
                 if sys.platform == "win32":
                     os.unlink(target_zip)
-                os.rename(target_temp,target_zip)
+                    time.sleep(0.01)
+                really_rename(target_temp,target_zip)
         else:
             raise ValueError("invalid command: " + cmd)
     finally:
+        if stream is not None:
+            if stream not in (sys.stdin,sys.stdout,):
+                stream.close()
         if opts.zipped:
-            shutil.rmtree(workdir)
+            really_rmtree(workdir)
  
 
 if __name__ == "__main__":
