@@ -6,7 +6,7 @@
 
 This module provides the infrastructure for spawning a stand-alone "helper app"
 to install updates with root privileges.  The class "SudoProxy" provides a
-proxy to the methods of an object via a root-preivileged helper process.
+proxy to the methods of an object via a root-privileged helper process.
 
 Example:
 
@@ -108,6 +108,8 @@ class SudoProxy(object):
 
     def start(self):
         (self.proc,self.pipe) = spawn_sudo(self)
+        if self.proc.poll() is not None:
+            raise RuntimeError("sudo helper process terminated unexpectedly")
         if self.pipe.read() != b("READY"):
             self.close()
             raise RuntimeError("failed to spawn helper app")
