@@ -29,7 +29,8 @@ import esky.sudo
 from esky import bdist_esky
 from esky.bdist_esky import Executable
 from esky.util import extract_zipfile, deep_extract_zipfile, get_platform, \
-                      ESKY_CONTROL_DIR, files_differ
+                      ESKY_CONTROL_DIR, files_differ, ESKY_APPDATA_DIR, \
+                      really_rmtree
 from esky.fstransact import FSTransaction
 
 try:
@@ -48,6 +49,10 @@ try:
     import cx_Freeze
 except ImportError:
     cx_Freeze = None
+try:
+    import pypy
+except ImportError:
+    pypy = None
 
 sys.path.append(os.path.dirname(__file__))
 
@@ -80,42 +85,62 @@ def setenv(key,value):
 class TestEsky(unittest.TestCase):
 
   if py2exe is not None:
+
     def test_esky_py2exe(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe"}})
+
     def test_esky_py2exe_bundle1(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
                                             "freezer_options": {
                                               "bundle_files": 1}}})
+
     def test_esky_py2exe_bundle2(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
                                             "freezer_options": {
                                               "bundle_files": 2}}})
+
     def test_esky_py2exe_bundle3(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
                                             "freezer_options": {
                                               "bundle_files": 3}}})
-    if sys.platform == "win32":
-        def test_esky_py2exe_nocustomchainload(self):
-            with setenv("ESKY_NO_CUSTOM_CHAINLOAD","1"):
-               bscode = "_chainload = _orig_chainload\nbootstrap()"
-               self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
-                                                   "bootstrap_code":bscode}})
+
+    def test_esky_py2exe_nocustomchainload(self):
+        with setenv("ESKY_NO_CUSTOM_CHAINLOAD","1"):
+           bscode = "_chainload = _orig_chainload\nbootstrap()"
+           self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
+                                               "bootstrap_code":bscode}})
+
     if esky.sudo.can_get_root():
         def test_esky_py2exe_needsroot(self):
             with setenv("ESKY_NEEDSROOT","1"):
                self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe"}})
 
+    if pypy is not None:
+        def test_esky_py2exe_pypy(self):
+            self._run_eskytester({"bdist_esky":{"freezer_module":"py2exe",
+                                                "compile_bootstrap_exes":1}})
+
+
   if py2app is not None:
+
     def test_esky_py2app(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"py2app"}})
+
     if esky.sudo.can_get_root():
         def test_esky_py2app_needsroot(self):
             with setenv("ESKY_NEEDSROOT","1"):
                 self._run_eskytester({"bdist_esky":{"freezer_module":"py2app"}})
 
+    if pypy is not None:
+        def test_esky_py2app_pypy(self):
+            self._run_eskytester({"bdist_esky":{"freezer_module":"py2app",
+                                                "compile_bootstrap_exes":1}})
+
   if bbfreeze is not None:
+
     def test_esky_bbfreeze(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"bbfreeze"}})
+
     if sys.platform == "win32":
         def test_esky_bbfreeze_nocustomchainload(self):
             with setenv("ESKY_NO_CUSTOM_CHAINLOAD","1"):
@@ -127,20 +152,34 @@ class TestEsky(unittest.TestCase):
             with setenv("ESKY_NEEDSROOT","1"):
                 self._run_eskytester({"bdist_esky":{"freezer_module":"bbfreeze"}})
 
+    if pypy is not None:
+        def test_esky_bbfreeze_pypy(self):
+            self._run_eskytester({"bdist_esky":{"freezer_module":"bbfreeze",
+                                                "compile_bootstrap_exes":1}})
+
   if cx_Freeze is not None:
+
     def test_esky_cxfreeze(self):
         self._run_eskytester({"bdist_esky":{"freezer_module":"cxfreeze"}})
+
     if sys.platform == "win32":
         def test_esky_cxfreeze_nocustomchainload(self):
             with setenv("ESKY_NO_CUSTOM_CHAINLOAD","1"):
                bscode = ["_chainload = _orig_chainload",None]
                self._run_eskytester({"bdist_esky":{"freezer_module":"cxfreeze",
                                                    "bootstrap_code":bscode}})
+
     if esky.sudo.can_get_root():
         def test_esky_cxfreeze_needsroot(self):
             with setenv("ESKY_NEEDSROOT","1"):
                 self._run_eskytester({"bdist_esky":{"freezer_module":"cxfreeze"}})
 
+    if pypy is not None:
+        def test_esky_cxfreeze_pypy(self):
+            with setenv("ESKY_NO_CUSTOM_CHAINLOAD","1"):
+              self._run_eskytester({"bdist_esky":{"freezer_module":"cxfreeze",
+                                                 "compile_bootstrap_exes":1}})
+ 
 
   def _run_eskytester(self,options):
     """Build and run the eskytester app using the given distutils options.
@@ -149,8 +188,13 @@ class TestEsky(unittest.TestCase):
     sequence of tests performed range across "script1.py" to "script3.py".
     """
     olddir = os.path.abspath(os.curdir)
+#    tdir = os.path.join(os.path.dirname(__file__),"DIST")
+#    if os.path.exists(tdir):
+#        really_rmtree(tdir)
+#    os.mkdir(tdir)
     tdir = tempfile.mkdtemp()
     server = None
+    script2 = None
     try:
         options.setdefault("build",{})["build_base"] = os.path.join(tdir,"build")
         options.setdefault("bdist",{})["dist_dir"] = os.path.join(tdir,"dist")
@@ -162,9 +206,6 @@ class TestEsky(unittest.TestCase):
         esky_root = dirname(dirname(dirname(__file__)))
         os.chdir(tdir)
         shutil.copytree(os.path.join(esky_root,"esky","tests","eskytester"),"eskytester")
-        #  Clean up after previous test runs.
-        if os.path.isdir(deploydir):
-            shutil.rmtree(deploydir)
         dir_util._path_created.clear()
         #  Build three increasing versions of the test package.
         #  Version 0.2 will include a bundled MSVCRT on win32.
@@ -205,11 +246,12 @@ class TestEsky(unittest.TestCase):
         extract_zipfile(zfname,deploydir)
         #  Run the scripts in order.
         if options["bdist_esky"]["freezer_module"] == "py2app":
-            appdir = os.listdir(deploydir)[0]
-            cmd1 = os.path.join(deploydir,appdir,"Contents","MacOS","script1")
-            cmd2 = os.path.join(deploydir,appdir,"Contents","MacOS","script2")
-            cmd3 = os.path.join(deploydir,appdir,"Contents","MacOS","script3")
+            appdir = os.path.join(deploydir,os.listdir(deploydir)[0])
+            cmd1 = os.path.join(appdir,"Contents","MacOS","script1")
+            cmd2 = os.path.join(appdir,"Contents","MacOS","script2")
+            cmd3 = os.path.join(appdir,"Contents","MacOS","script3")
         else:
+            appdir = deploydir
             if sys.platform == "win32":
                 cmd1 = os.path.join(deploydir,"script1.exe")
                 cmd2 = os.path.join(deploydir,"script2.exe")
@@ -219,36 +261,27 @@ class TestEsky(unittest.TestCase):
                 cmd2 = os.path.join(deploydir,"script2")
                 cmd3 = os.path.join(deploydir,"script3")
         print "spawning eskytester script1", options["bdist_esky"]["freezer_module"]
-        p = subprocess.Popen(cmd1)#,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
-        #(stdout,_) = p.communicate()
-        #sys.stdout.write(stdout.decode())
-        p.wait()
-        assert p.returncode == 0
-        assert os.path.exists("tests-completed")
-        os.unlink("tests-completed")
+        os.unlink(os.path.join(tdir,"dist","eskytester-0.1.%s.zip"%(platform,)))
+        p = subprocess.Popen(cmd1)
+        assert p.wait() == 0
+        os.unlink(os.path.join(appdir,"tests-completed"))
         print "spawning eskytester script2"
-        p = subprocess.Popen(cmd2,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
-        (stdout,_) = p.communicate()
-        sys.stdout.write(stdout.decode())
-        assert p.returncode == 0
-        assert os.path.exists("tests-completed")
-        os.unlink("tests-completed")
+        os.unlink(os.path.join(tdir,"dist","eskytester-0.2.%s.zip"%(platform,)))
+        p = subprocess.Popen(cmd2)
+        assert p.wait() == 0
+        os.unlink(os.path.join(appdir,"tests-completed"))
         print "spawning eskytester script3"
-        p = subprocess.Popen(cmd3,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
-        (stdout,_) = p.communicate()
-        sys.stdout.write(stdout.decode())
-        assert p.returncode == 0
-        assert os.path.exists("tests-completed")
-        os.unlink("tests-completed")
+        p = subprocess.Popen(cmd3)
+        assert p.wait() == 0
+        os.unlink(os.path.join(appdir,"tests-completed"))
     finally:
+        if script2:
+            script2.script[1].close()
         os.chdir(olddir)
-        for i in xrange(10):
-            try:
-                shutil.rmtree(tdir)
-            except EnvironmentError:
-                time.sleep(2)
-            else:
-                break
+        if sys.platform == "win32":
+           # wait for the cleanup-at-exit pocess to finish
+           time.sleep(4)
+        really_rmtree(tdir)
         if server:
             server.shutdown()
  
@@ -257,8 +290,8 @@ class TestEsky(unittest.TestCase):
     platform = get_platform()
     appdir = tempfile.mkdtemp()
     try: 
-        vdir = os.path.join(appdir,"testapp-0.1.%s" % (platform,))
-        os.mkdir(vdir)
+        vdir = os.path.join(appdir,ESKY_APPDATA_DIR,"testapp-0.1.%s" % (platform,))
+        os.makedirs(vdir)
         os.mkdir(os.path.join(vdir,ESKY_CONTROL_DIR))
         open(os.path.join(vdir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt"),"wb").close()
         e1 = esky.Esky(appdir,"http://example.com/downloads/")
@@ -300,9 +333,8 @@ class TestEsky(unittest.TestCase):
     """Test that breaking the lock on an Esky works correctly."""
     appdir = tempfile.mkdtemp()
     try: 
-        os.mkdir(os.path.join(appdir,"testapp-0.1"))
-        os.mkdir(os.path.join(appdir,"testapp-0.1",ESKY_CONTROL_DIR))
-        open(os.path.join(appdir,"testapp-0.1",ESKY_CONTROL_DIR,"bootstrap-manifest.txt"),"wb").close()
+        os.makedirs(os.path.join(appdir,ESKY_APPDATA_DIR,"testapp-0.1",ESKY_CONTROL_DIR))
+        open(os.path.join(appdir,ESKY_APPDATA_DIR,"testapp-0.1",ESKY_CONTROL_DIR,"bootstrap-manifest.txt"),"wb").close()
         e1 = esky.Esky(appdir,"http://example.com/downloads/")
         e2 = esky.Esky(appdir,"http://example.com/downloads/")
         trigger1 = threading.Event(); trigger2 = threading.Event()
@@ -644,14 +676,8 @@ class TestPatch(unittest.TestCase):
 
     def _extract(self,filename,dest):
         dest = os.path.join(self.workdir,dest)
-        for i in xrange(10):
-            if os.path.exists(dest):
-                try:
-                    shutil.rmtree(dest)
-                except EnvironmentError:
-                    time.sleep(1)
-                else:
-                    break
+        if os.path.exists(dest):
+            really_rmtree(dest)
         f = tarfile.open(os.path.join(self.tfdir,filename),"r:gz")
         try:
             f.extractall(dest)
