@@ -46,7 +46,6 @@ def lazy_import(func):
     return _LazyImport(func.func_name,func,namespace)
 
 
-
 class _LazyImport(object):
     """Class representing a lazy import."""
 
@@ -91,6 +90,11 @@ def shutil():
     import shutil
     return shutil
 
+@lazy_import
+def time():
+    import time
+    return time
+
 @lazy_import
 def re():
     import re
@@ -128,7 +132,8 @@ from esky.bootstrap import get_best_version, get_all_versions,\
                            is_uninstalled_version_dir,\
                            split_app_version, join_app_version, parse_version,\
                            get_original_filename, lock_version_dir,\
-                           unlock_version_dir, fcntl, ESKY_CONTROL_DIR
+                           unlock_version_dir, fcntl, ESKY_CONTROL_DIR,\
+                           ESKY_APPDATA_DIR
 
 
 def files_differ(file1,file2,start=0,stop=None):
@@ -200,11 +205,19 @@ def common_prefix(iterables):
         prefix = prefix[:count]
     return prefix
 
+
 def appdir_from_executable(exepath):
     """Find the top-level application directory, given sys.executable."""
+    #  The standard layout is <appdir>/ESKY_APPDATA_DIR/<vdir>/<exepath>.
+    #  Stripping of <exepath> is done by _bs_appdir_from_executable.
     vdir = _bs_appdir_from_executable(exepath)
     appdir = os.path.dirname(vdir)
-    if os.path.exists(os.path.join(appdir,ESKY_CONTROL_DIR,"bootstrap-manifest.txt")):
+    # TODO: remove compatability hook for ESKY_APPDATA_DIR=""
+    if ESKY_APPDATA_DIR and os.path.basename(appdir) == ESKY_APPDATA_DIR:
+        appdir = os.path.dirname(appdir)
+    #  On OSX we sometimes need to strip an additional directory since the
+    #  app can be contained in an <appname>.app directory.
+    if is_version_dir(appdir):
         appdir = os.path.dirname(appdir)
     return appdir
 
@@ -261,6 +274,7 @@ def zipfile_common_prefix_dir(source):
     else:
         return ""
 
+
 def deep_extract_zipfile(source,target,name_filter=None):
     """Extract the deep contents of a zipfile into a target directory.
 
@@ -421,3 +435,60 @@ def is_locked_version_dir(vdir):
         finally:
             f.close()
 
+
+def really_rename(source,target):
+    """Like os.rename, but try to work around some win32 wierdness.
+
+    Every so often windows likes to throw a spurious error about not being
+    able to rename something; if we sleep for a brief period and try
+    again it seems to get over it.
+    """
+    if sys.platform != "win32":
+        os.rename(source,target)
+    else:
+        for _ in xrange(3):
+            try:
+                os.rename(source,target)
+            except WindowsError, e:
+                if e.errno not in (errno.EACCES,):
+                    raise
+                time.sleep(0.01)
+            else:
+                break
+        else:
+            os.rename(source,target)
+
+
+def really_rmtree(path):
+    """Like shutil.rmtree, but try to work around some win32 wierdness.
+
+    Every so often windows likes to throw a spurious error about not being
+    able to remove a directory - like claiming it still contains files after
+    we just deleted all the files in the directory.  If we sleep for a brief
+    period and try again it seems to get over it.
+    """
+    if sys.platform != "win32":
+        shutil.rmtree(path)
+    else:
+        #  If it's going to error out legitimately, let it do so.
+        if not os.path.exists(path):
+            shutil.rmtree(path)
+        #  This is a little retry loop that catches troublesome errors.
+        for _ in xrange(3):
+            try:
+                shutil.rmtree(path)
+            except WindowsError, e:
+                if e.errno in (errno.ENOTEMPTY,errno.EACCES,):
+                    time.sleep(0.01)
+                elif e.errno == errno.ENOENT:
+                    if not os.path.exists(path):
+                        return
+                    time.sleep(0.01)
+                else:
+                    raise
+            else:
+                break
+        else:
+            shutil.rmtree(path)
+
+
